@@ -1,4 +1,4 @@
-/* $Id: deliver-smtp.c,v 1.49 2007/03/28 19:59:57 nicm Exp $ */
+/* $Id: deliver-smtp.c,v 1.53 2007/08/18 12:19:43 nicm Exp $ */
 
 /*
  * Copyright (c) 2006 Nicholas Marriott <nicm@users.sourceforge.net>
@@ -75,7 +75,7 @@ deliver_smtp_deliver(struct deliver_ctx *dctx, struct actitem *ti)
 	size_t		 		 len, llen;
 
 	io = connectproxy(&data->server,
-	    conf.proxy, IO_CRLF, conf.timeout, &cause);
+	    conf.verify_certs, conf.proxy, IO_CRLF, conf.timeout, &cause);
 	if (io == NULL) {
 		log_warnx("%s: %s", a->name, cause);
 		xfree(cause);
@@ -102,7 +102,8 @@ deliver_smtp_deliver(struct deliver_ctx *dctx, struct actitem *ti)
 	line = cause = NULL;
 	done = 0;
 	do {
-		switch (io_pollline2(io, &line, &lbuf, &llen, &cause)) {
+		switch (io_pollline2(io,
+		    &line, &lbuf, &llen, conf.timeout, &cause)) {
 		case 0:
 			cause = xstrdup("connection unexpectedly closed");
 			goto error;
@@ -141,18 +142,19 @@ deliver_smtp_deliver(struct deliver_ctx *dctx, struct actitem *ti)
 				goto error;
 			line_init(m, &ptr, &len);
 			while (ptr != NULL) {
-				io_write(io, ptr, len - 1);
+				if (len > 1)
+					io_write(io, ptr, len - 1);
 				io_writeline(io, NULL);
 
-				/* update if necessary */
-				if (io_update(io, &cause) != 1)
+				/* Update if necessary. */
+				if (io_update(io, conf.timeout, &cause) != 1)
 					goto error;
 
 				line_next(m, &ptr, &len);
 			}
 			state = SMTP_DONE;
 			io_writeline(io, ".");
-			io_flush(io, NULL);
+			io_flush(io, conf.timeout, NULL);
 			break;
 		case SMTP_DONE:
 			if (code != 250)
@@ -190,7 +192,7 @@ error:
 		log_warnx("%s: unexpected response: %s", a->name, line);
 
 	io_writeline(io, "QUIT");
-	io_flush(io, NULL);
+	io_flush(io, conf.timeout, NULL);
 
 	xfree(lbuf);
 	if (from != NULL)
