@@ -1,4 +1,4 @@
-/* $Id: fetch-maildir.c,v 1.64 2007/03/26 16:01:38 nicm Exp $ */
+/* $Id: fetch-maildir.c,v 1.84 2007/08/30 10:45:06 nicm Exp $ */
 
 /*
  * Copyright (c) 2006 Nicholas Marriott <nicm@users.sourceforge.net>
@@ -31,28 +31,32 @@
 #include "fdm.h"
 #include "fetch.h"
 
-int	 fetch_maildir_start(struct account *, int *);
-int	 fetch_maildir_finish(struct account *, int);
-int	 fetch_maildir_poll(struct account *, u_int *);
-int	 fetch_maildir_fetch(struct account *, struct mail *);
-int	 fetch_maildir_done(struct account *, struct mail *);
-void	 fetch_maildir_desc(struct account *, char *, size_t);
+int	fetch_maildir_commit(struct account *, struct mail *);
+void	fetch_maildir_abort(struct account *);
+u_int	fetch_maildir_total(struct account *);
+void	fetch_maildir_desc(struct account *, char *, size_t);
 
-void	 fetch_maildir_free(void *);
+void	fetch_maildir_free(void *);
 
-int	 fetch_maildir_makepaths(struct account *);
-void	 fetch_maildir_freepaths(struct account *);
+int	fetch_maildir_makepaths(struct account *);
+void	fetch_maildir_freepaths(struct account *);
+
+int	fetch_maildir_poll(struct account *);
+
+int	fetch_maildir_state_init(struct account *, struct fetch_ctx *);
+int	fetch_maildir_state_build(struct account *, struct fetch_ctx *);
+int	fetch_maildir_state_next(struct account *, struct fetch_ctx *);
+int	fetch_maildir_state_open(struct account *, struct fetch_ctx *);
+int	fetch_maildir_state_mail(struct account *, struct fetch_ctx *);
 
 struct fetch fetch_maildir = {
 	"maildir",
-	{ NULL, NULL },
-	fetch_maildir_start,
+	fetch_maildir_state_init,
+
 	NULL,
-	fetch_maildir_poll,
-	fetch_maildir_fetch,
-	NULL,
-	fetch_maildir_done,
-	fetch_maildir_finish,
+	fetch_maildir_commit,
+	fetch_maildir_abort,
+	fetch_maildir_total,
 	fetch_maildir_desc
 };
 
@@ -78,17 +82,17 @@ fetch_maildir_makepaths(struct account *a)
 	ARRAY_INIT(data->paths);
 
 	for (i = 0; i < ARRAY_LENGTH(data->maildirs); i++) {
-		path = ARRAY_ITEM(data->maildirs, i, char *);
+		path = ARRAY_ITEM(data->maildirs, i);
 		if (glob(path, GLOB_BRACE|GLOB_NOCHECK, NULL, &g) != 0) {
 			log_warn("%s: glob(\"%s\")", a->name, path);
 			goto error;
 		}
 
 		if (g.gl_pathc < 1)
-			fatalx("negative or zero number of paths");
+			fatalx("glob returned garbage");
 		for (j = 0; j < (u_int) g.gl_pathc; j++) {
 			xasprintf(&path, "%s/cur", g.gl_pathv[j]);
-			ARRAY_ADD(data->paths, path, char *);
+			ARRAY_ADD(data->paths, path);
 			if (stat(path, &sb) != 0) {
 				log_warn("%s: %s", a->name, path);
 				goto error;
@@ -100,7 +104,7 @@ fetch_maildir_makepaths(struct account *a)
 			}
 
 			xasprintf(&path, "%s/new", g.gl_pathv[j]);
-			ARRAY_ADD(data->paths, path, char *);
+			ARRAY_ADD(data->paths, path);
 			if (stat(path, &sb) != 0) {
 				log_warn("%s", path);
 				goto error;
@@ -111,13 +115,15 @@ fetch_maildir_makepaths(struct account *a)
 				goto error;
 			}
 		}
+
+		globfree(&g);
 	}
 
 	return (0);
 
 error:
 	fetch_maildir_freepaths(a);
-	return (1);
+	return (-1);
 }
 
 /* Free the array. */
@@ -128,44 +134,14 @@ fetch_maildir_freepaths(struct account *a)
 	u_int			 	 i;
 
 	for (i = 0; i < ARRAY_LENGTH(data->paths); i++)
-		xfree(ARRAY_ITEM(data->paths, i, char *));
+		xfree(ARRAY_ITEM(data->paths, i));
 
 	ARRAY_FREEALL(data->paths);
 }
 
+/* Count maildir total. */
 int
-fetch_maildir_start(struct account *a, unused int *total)
-{
-	struct fetch_maildir_data	*data = a->data;
-
-	data->dirp = NULL;
-
-	data->path = NULL;
-
-	if (fetch_maildir_makepaths(a) != 0)
-		return (FETCH_ERROR);
-	data->index = 0;
-
-	return (FETCH_SUCCESS);
-}
-
-int
-fetch_maildir_finish(struct account *a, unused int aborted)
-{
-	struct fetch_maildir_data	*data = a->data;
-
-	fetch_maildir_freepaths(a);
-
-	if (data->dirp != NULL && closedir(data->dirp) != 0) {
-		log_warn("%s: %s: closedir", a->name, data->path);
-		return (FETCH_ERROR);
-	}
-
-	return (FETCH_SUCCESS);
-}
-
-int
-fetch_maildir_poll(struct account *a, u_int *n)
+fetch_maildir_poll(struct account *a)
 {
 	struct fetch_maildir_data	*data = a->data;
 	u_int				 i;
@@ -174,167 +150,239 @@ fetch_maildir_poll(struct account *a, u_int *n)
 	struct dirent			*dp;
 	struct stat			 sb;
 
-	*n = 0;
+	data->total = 0;
 	for (i = 0; i < ARRAY_LENGTH(data->paths); i++) {
-		path = ARRAY_ITEM(data->paths, i, char *);
+		path = ARRAY_ITEM(data->paths, i);
 
 		log_debug("%s: trying path: %s", a->name, path);
 		if ((dirp = opendir(path)) == NULL) {
 			log_warn("%s: %s: opendir", a->name, path);
-			return (FETCH_ERROR);
+			return (-1);
 		}
 
 		while ((dp = readdir(dirp)) != NULL) {
 			if (dp->d_type == DT_REG) {
-				(*n)++;
+				data->total++;
 				continue;
 			}
 			if (dp->d_type != DT_UNKNOWN)
 				continue;
 
-			if (printpath(entry, sizeof entry, "%s/%s", path,
+			if (mkpath(entry, sizeof entry, "%s/%s", path,
 			    dp->d_name) != 0) {
 				log_warn("%s: %s: printpath", a->name, path);
 				closedir(dirp);
-				return (FETCH_ERROR);
+				return (-1);
 			}
 
 			if (stat(entry, &sb) != 0) {
 				log_warn("%s: %s: stat", a->name, entry);
 				closedir(dirp);
-				return (FETCH_ERROR);
+				return (-1);
 			}
 			if (!S_ISREG(sb.st_mode))
 				continue;
 
-			(*n)++;
+			data->total++;
 		}
 
 		if (closedir(dirp) != 0) {
 			log_warn("%s: %s: closedir", a->name, path);
-			return (FETCH_ERROR);
+			return (-1);
 		}
 	}
 
-	return (FETCH_SUCCESS);
+	return (0);
 }
 
+/* Commit mail. */
 int
-fetch_maildir_fetch(struct account *a, struct mail *m)
+fetch_maildir_commit(struct account *a, struct mail *m)
 {
-	struct fetch_maildir_data	*data = a->data;
 	struct fetch_maildir_mail	*aux;
-	struct dirent			*dp;
-	char	       			*ptr;
-	struct stat			 sb;
-	int				 fd;
 
-restart:
-	if (data->dirp == NULL) {
-		data->path = ARRAY_ITEM(data->paths, data->index, char *);
-
-		/* make the maildir name for the tag */
-		strlcpy(data->maildir,
-		    xbasename(xdirname(data->path)), sizeof data->maildir);
-
-		log_debug2("%s: trying path: %s", a->name, data->path);
-		if ((data->dirp = opendir(data->path)) == NULL) {
-			log_warn("%s: %s: opendir", a->name, data->path);
-			return (FETCH_ERROR);
-		}
-	}
-
-	do {
-		dp = readdir(data->dirp);
-		if (dp == NULL) {
-			if (closedir(data->dirp) != 0) {
-				log_warn("%s: %s: closedir", a->name,
-				    data->path);
-				return (FETCH_ERROR);
-			}
-			data->dirp = NULL;
-
-			data->index++;
-			if (data->index == ARRAY_LENGTH(data->paths))
-				return (FETCH_COMPLETE);
-			goto restart;
-		}
-
-		if (printpath(data->entry, sizeof data->entry, "%s/%s",
-		    data->path, dp->d_name) != 0) {
-			log_warn("%s: %s: printpath", a->name, data->path);
-			return (FETCH_ERROR);
-		}
-		if (stat(data->entry, &sb) != 0) {
-			log_warn("%s: %s: stat", a->name, data->entry);
-			return (FETCH_ERROR);
-		}
-	} while (!S_ISREG(sb.st_mode));
-
-	log_debug2("%s: reading mail from: %s", a->name, data->entry);
-	if (sb.st_size <= 0)
-		return (FETCH_EMPTY);
-	if ((uintmax_t) sb.st_size > SIZE_MAX)
-		return (FETCH_OVERSIZE);
-	if ((uintmax_t) sb.st_size > conf.max_size)
-		return (FETCH_OVERSIZE);
-
-	if ((fd = open(data->entry, O_RDONLY, 0)) < 0) {
-		log_warn("%s: %s: open", a->name, data->entry);
-		return (FETCH_ERROR);
-	}
-
-	if (mail_open(m, IO_ROUND(sb.st_size)) != 0) {
-		log_warn("%s: failed to create mail", a->name);
-		return (FETCH_ERROR);
-	}
-	default_tags(&m->tags, data->maildir, a);
-	add_tag(&m->tags, "maildir", "%s", data->maildir);
-
-	aux = xmalloc(sizeof *aux);
-	strlcpy(aux->path, data->entry, sizeof aux->path);
-	m->auxdata = aux;
-	m->auxfree = fetch_maildir_free;
-
-	log_debug2("%s: reading %ju bytes", a->name, (uintmax_t) sb.st_size);
-	if (read(fd, m->data, sb.st_size) != sb.st_size) {
-		close(fd);
-		log_warn("%s: %s: read", a->name, data->entry);
-		return (FETCH_ERROR);
-	}
-	close(fd);
-
-	/* find the body */
-	m->body = -1;
-	ptr = m->data;
-	while ((ptr = memchr(ptr, '\n', (m->data + m->size) - ptr)) != NULL) {
-		ptr++;
-		if (ptr < (m->data + m->size) && *ptr == '\n') {
-			ptr++;
-			if (ptr != (m->data + m->size))
-				m->body = ptr - m->data;
-			break;
-		}
-	}
-	m->size = sb.st_size;
-
-	return (FETCH_SUCCESS);
-}
-
-int
-fetch_maildir_done(struct account *a, struct mail *m)
-{
-	struct fetch_maildir_mail	*aux = m->auxdata;
-
-	if (m->decision == DECISION_KEEP)
-		return (FETCH_SUCCESS);
-
-	if (unlink(aux->path) != 0) {
+	aux = m->auxdata;
+	if (m->decision == DECISION_DROP && unlink(aux->path) != 0) {
 		log_warn("%s: %s: unlink", a->name, aux->path);
 		return (FETCH_ERROR);
 	}
 
-	return (FETCH_SUCCESS);
+	return (FETCH_AGAIN);
+}
+
+/* Abort fetch. */
+void
+fetch_maildir_abort(struct account *a)
+{
+	struct fetch_maildir_data	*data = a->data;
+
+	if (data->dirp != NULL)
+		closedir(data->dirp);
+	fetch_maildir_freepaths(a);
+}
+
+/* Return total mails. */
+u_int
+fetch_maildir_total(struct account *a)
+{
+	struct fetch_maildir_data	*data = a->data;
+
+	return (data->total);
+}
+
+/* Initialise maildir fetch context. */
+int
+fetch_maildir_state_init(struct account *a, struct fetch_ctx *fctx)
+{
+	struct fetch_maildir_data	*data = a->data;
+
+	if (fetch_maildir_makepaths(a) != 0)
+		return (FETCH_ERROR);
+	if (ARRAY_EMPTY(data->paths)) {
+		log_warnx("%s: no maildirs found", a->name);
+		return (-1);
+	}
+
+	data->index = 0;
+	data->dirp = NULL;
+
+	/* Poll counts mails and exits. */
+	if (fctx->flags & FETCH_POLL) {
+		if (fetch_maildir_poll(a) != 0)
+			return (FETCH_ERROR);
+		fetch_maildir_freepaths(a);
+		return (FETCH_EXIT);
+	}
+
+	fctx->state = fetch_maildir_state_open;
+	return (0);
+}
+
+/* Next state. Move to next path. */
+int
+fetch_maildir_state_next(struct account *a, struct fetch_ctx *fctx)
+{
+	struct fetch_maildir_data	*data = a->data;
+
+	if (data->index <= ARRAY_LENGTH(data->paths))
+		data->index++;
+
+	if (data->index == ARRAY_LENGTH(data->paths)) {
+		fetch_maildir_freepaths(a);
+		return (FETCH_EXIT);
+	}
+
+	fctx->state = fetch_maildir_state_open;
+	return (FETCH_AGAIN);
+}
+
+/* Open state. */
+int
+fetch_maildir_state_open(struct account *a, unused struct fetch_ctx *fctx)
+{
+	struct fetch_maildir_data	*data = a->data;
+	char				*path;
+
+ 	path = ARRAY_ITEM(data->paths, data->index);
+
+	/* Open the directory. */
+	log_debug2("%s: trying path: %s", a->name, path);
+	if ((data->dirp = opendir(path)) == NULL) {
+		log_warn("%s: %s: opendir", a->name, path);
+		return (FETCH_ERROR);
+	}
+
+	fctx->state = fetch_maildir_state_mail;
+	return (FETCH_AGAIN);
+}
+
+/* Mail state. Find and read mail file. */
+int
+fetch_maildir_state_mail(struct account *a, struct fetch_ctx *fctx)
+{
+	struct fetch_maildir_data	*data = a->data;
+	struct mail			*m = fctx->mail;
+	struct fetch_maildir_mail	*aux;
+	struct dirent			*dp;
+	char	       			*path, *maildir, name[MAXPATHLEN];
+	struct stat			 sb;
+	uintmax_t			 size;
+	int				 fd;
+	ssize_t				 n;
+
+ 	path = ARRAY_ITEM(data->paths, data->index);
+
+restart:
+	/* Read the next dir entry. */
+	dp = readdir(data->dirp);
+	if (dp == NULL) {
+		if (closedir(data->dirp) != 0) {
+			log_warn("%s: %s: closedir", a->name, path);
+			return (FETCH_ERROR);
+		}
+		data->dirp = NULL;
+
+		fctx->state = fetch_maildir_state_next;
+		return (FETCH_AGAIN);
+	}
+
+	if (mkpath(name, sizeof name, "%s/%s", path, dp->d_name) != 0) {
+		log_warn("%s: %s: printpath", a->name, path);
+		return (FETCH_ERROR);
+	}
+	if (stat(name, &sb) != 0) {
+		log_warn("%s: %s: stat", a->name, name);
+		return (FETCH_ERROR);
+	}
+	if (!S_ISREG(sb.st_mode))
+		goto restart;
+
+	/* Open the mail. */
+	if (mail_open(m, sb.st_size) != 0) {
+		log_warn("%s: failed to create mail", a->name);
+		return (FETCH_ERROR);
+	}
+
+	/* Got a valid entry, start reading it. */
+	log_debug2("%s: reading mail from: %s", a->name, name);
+	size = sb.st_size;
+	if (sb.st_size <= 0) {
+		m->size = 0;
+		return (FETCH_MAIL);
+	} else if (size > SIZE_MAX || size > conf.max_size) {
+		m->size = SIZE_MAX;
+		return (FETCH_MAIL);
+	}
+
+	/* Open the file. */
+	if ((fd = open(name, O_RDONLY, 0)) == -1) {
+		log_warn("%s: %s: open", a->name, name);
+		return (FETCH_ERROR);
+	}
+
+	/* Add the tags. */
+	maildir = xbasename(xdirname(path));
+	default_tags(&m->tags, maildir);
+	add_tag(&m->tags, "maildir", "%s", maildir);
+
+	/* Add aux data. */
+	aux = xmalloc(sizeof *aux);
+	strlcpy(aux->path, name, sizeof aux->path);
+	m->auxdata = aux;
+	m->auxfree = fetch_maildir_free;
+
+	/* Read the mail. */
+	if ((n = read(fd, m->data, size)) == -1 || (size_t) n != size) {
+		close(fd);
+		log_warn("%s: %s: read", a->name, name);
+		return (FETCH_ERROR);
+	}
+	close(fd);
+
+	log_debug2("%s: read %ju bytes", a->name, size);
+	m->size = size;
+
+	return (FETCH_MAIL);
 }
 
 void
@@ -343,7 +391,7 @@ fetch_maildir_desc(struct account *a, char *buf, size_t len)
 	struct fetch_maildir_data	*data = a->data;
 	char				*maildirs;
 
-	maildirs = fmt_strings("maildirs ", data->maildirs);
+	maildirs = fmt_strings("maildir ", data->maildirs);
 	strlcpy(buf, maildirs, len);
 	xfree(maildirs);
 }
