@@ -1,4 +1,4 @@
-/* $Id: replace.c,v 1.36 2007/03/26 21:27:50 nicm Exp $ */
+/* $Id: replace.c,v 1.46 2007/09/25 17:45:38 nicm Exp $ */
 
 /*
  * Copyright (c) 2006 Nicholas Marriott <nicm@users.sourceforge.net>
@@ -85,6 +85,9 @@ static const char *aliases[] = {
 	NULL, 		/* Z */
 };
 
+char		*replace(char *, struct strb *, struct mail *, struct rmlist *);
+const char 	*submatch(char, struct mail *, struct rmlist *, size_t *);
+
 void printflike3
 add_tag(struct strb **tags, const char *key, const char *value, ...)
 {
@@ -120,8 +123,9 @@ match_tag(struct strb *tags, const char *pattern)
 }
 
 void
-default_tags(struct strb **tags, char *src, struct account *a)
+default_tags(struct strb **tags, const char *src)
 {
+	char		 rtime[128];
 	struct tm	*tm;
 	time_t		 t;
 
@@ -132,21 +136,35 @@ default_tags(struct strb **tags, char *src, struct account *a)
 
 	if (src != NULL)
 		add_tag(tags, "source", "%s", src);
-	if (a != NULL)
-		add_tag(tags, "account", "%s", a->name);
+
+	if (conf.info.host != NULL)
+		add_tag(tags, "hostname", "%s", conf.info.host);
 
 	t = time(NULL);
 	if ((tm = localtime(&t)) != NULL) {
+		/*
+		 * Okay, in a struct tm, everything is zero-based (including
+		 * month!) except day of the month which is one-based.
+		 *
+		 * To make thing clearer, strftime(3) measures everything as
+		 * you would expect... except that day of the week runs from
+		 * 0-6 but day of the year runs from 1-366.
+		 *
+		 * Fun fun fun.
+		 */
 		add_tag(tags, "hour", "%.2d", tm->tm_hour);
 		add_tag(tags, "minute", "%.2d", tm->tm_min);
 		add_tag(tags, "second", "%.2d", tm->tm_sec);
 		add_tag(tags, "day", "%.2d", tm->tm_mday);
-		add_tag(tags, "month", "%.2d", tm->tm_mon);
+		add_tag(tags, "month", "%.2d", tm->tm_mon + 1);
 		add_tag(tags, "year", "%.4d", 1900 + tm->tm_year);
+		add_tag(tags, "year2", "%.2d", tm->tm_year % 100);
 		add_tag(tags, "dayofweek", "%d", tm->tm_wday);
-		add_tag(tags, "dayofyear", "%.2d", tm->tm_yday);
-		add_tag(tags, "quarter", "%d", (tm->tm_mon - 1) / 3 + 1);
+		add_tag(tags, "dayofyear", "%.2d", tm->tm_yday + 1);
+		add_tag(tags, "quarter", "%d", tm->tm_mon / 3 + 1);
 	}
+	if (rfc822time(t, rtime, sizeof rtime) != NULL)
+		add_tag(tags, "rfc822date", "%s", rtime);
 }
 
 void
@@ -178,14 +196,29 @@ replacepath(struct replpath *rp, struct strb *tags, struct mail *m,
 	return (ss);
 }
 
+const char *
+submatch(char ch, struct mail *m, struct rmlist *rml, size_t *len)
+{
+	struct rm	*rm;
+
+	if (rml == NULL || !rml->valid || m == NULL)
+		return (NULL);
+
+	rm = &rml->list[((u_char) ch) - '0'];
+	if (!rm->valid)
+		return (NULL);
+
+	*len = rm->eo - rm->so;
+	return (m->data + rm->so);
+}
+
 char *
 replace(char *src, struct strb *tags, struct mail *m, struct rmlist *rml)
 {
-	char		*ptr, *tend;
 	const char	*tptr, *alias;
-	char		*dst, ch;
-	size_t	 	 off, len, tlen;
-	u_int		 idx;
+	char		*ptr, *tend, *dst, ch;
+	size_t	 	 i, off, len, tlen;
+	int		 strip;
 
 	if (src == NULL)
 		return (NULL);
@@ -196,6 +229,7 @@ replace(char *src, struct strb *tags, struct mail *m, struct rmlist *rml)
 	len = REPLBUFSIZE;
 	dst = xmalloc(len);
 
+	strip = 1;
 	for (ptr = src; *ptr != '\0'; ptr++) {
 		switch (*ptr) {
 		case '%':
@@ -221,6 +255,14 @@ replace(char *src, struct strb *tags, struct mail *m, struct rmlist *rml)
 				continue;
 			}
 			ptr++;
+
+			if (*ptr == ':') {
+				strip = 0;
+				ptr++;
+			}
+			if (ptr == tend)
+				continue;
+
 			*tend = '\0';
 			if ((tptr = find_tag(tags, ptr)) == NULL) {
 				*tend = ']';
@@ -228,26 +270,34 @@ replace(char *src, struct strb *tags, struct mail *m, struct rmlist *rml)
 				continue;
 			}
 			tlen = strlen(tptr);
-
 			*tend = ']';
+
 			ptr = tend;
 			break;
+		case ':':
+			ch = *++ptr;
+			if (ch >= '0' && ch <= '9') {
+				tptr = submatch(ch, m, rml, &tlen);
+				if (tptr == NULL)
+					continue;
+				strip = 0;
+				break;
+			}
+
+			ENSURE_FOR(dst, len, off, 1);
+			dst[off++] = ch;
+			continue;
 		default:
 			if (ch >= '0' && ch <= '9') {
-				if (rml == NULL || !rml->valid || m == NULL)
+				tptr = submatch(ch, m, rml, &tlen);
+				if (tptr == NULL)
 					continue;
-				idx = ((u_char) ch) - '0';
-				if (!rml->list[idx].valid)
-					continue;
-
-				tptr = m->base + rml->list[idx].so;
-				tlen = rml->list[idx].eo - rml->list[idx].so;
 				break;
 			}
 
 			alias = NULL;
-			if (ALIAS_IDX(ch) != -1)
-				alias = aliases[ALIAS_IDX(ch)];
+			if (ALIAS_IDX((u_char) ch) != -1)
+				alias = aliases[ALIAS_IDX((u_char) ch)];
 			if (alias == NULL)
 				continue;
 
@@ -257,9 +307,18 @@ replace(char *src, struct strb *tags, struct mail *m, struct rmlist *rml)
 			break;
 		}
 
+		if (tlen == 0)
+			continue;
 		ENSURE_FOR(dst, len, off, tlen);
-		memcpy(dst + off, tptr, tlen);
-		off += tlen;
+		if (!strip) {
+			memcpy(dst + off, tptr, tlen);
+			off += tlen;
+			continue;
+		}
+		for (i = 0; i < tlen; i++) {
+			if (strchr(conf.strip_chars, tptr[i]) == NULL)
+				dst[off++] = tptr[i];
+		}
 	}
 
 out:
